/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
* Version: NPL 1.1/GPL 2.0/LGPL 2.1
*
* The contents of this file are subject to the Netscape Public License
* Version 1.1 (the "License"); you may not use this file except in
* compliance with the License. You may obtain a copy of the License at
* http://www.mozilla.org/NPL/
*
* Software distributed under the License is distributed on an "AS IS" basis,
* WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
* for the specific language governing rights and limitations under the
* License.
*
* The Original Code is mozilla.org code.
*
* The Initial Developer of the Original Code is
* Netscape Communications Corporation.
* Portions created by the Initial Developer are Copyright (C) 2002
* the Initial Developer. All Rights Reserved.
*
* Contributor(s):
*    David Haas <haasd@cae.wisc.edu>
*
*
* Alternatively, the contents of this file may be used under the terms of
* either the GNU General Public License Version 2 or later (the "GPL"), or
* the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
* in which case the provisions of the GPL or the LGPL are applicable instead
* of those above. If you wish to allow use of your version of this file only
* under the terms of either the GPL or the LGPL, and not to allow others to
* use your version of this file under the terms of the NPL, indicate your
* decision by deleting the provisions above and replace them with the notice
* and other provisions required by the GPL or the LGPL. If you do not delete
* the provisions above, a recipient may use your version of this file under
* the terms of any one of the NPL, the GPL or the LGPL.
*
* ***** END LICENSE BLOCK ***** */

#import "Bookmark.h"
#import "BookmarkFolder.h"
#import "BookmarkManager.h"
#import "BookmarksClient.h"
#import "NSString+Utils.h"
#import "RunLoopMessenger.h"
#import "SiteIconProvider.h"

@interface Bookmark (Private)
-(void) setUpdateNotification:(unsigned)aNumber;
-(unsigned) updateNotification;
-(void) setStatusWithNSNumber:(NSNumber *)aNumber;
-(void) initializeLock;
-(void) startTimerInBackground:(NSTimer *)aTimer;
-(void) siteIconCheck:(NSNotification *)aNote;
-(void) urlLoadCheck:(NSNotification *)aNote;
-(void) invalidSchemeForUpdating;
-(void) doHTTPUpdateRequest:(NSURL *)aURL;
-(void) doFileUpdateRequest:(NSURL *)aURL;
-(void) interpretHTTPUpdateCode:(UInt32) aCode;
-(void) cleanupHTTPCheck:(NSTimer *)aTimer;
-(void) doUpdateNotificationTasks;

@end

// Constants - for update status checking
static const CFOptionFlags kNetworkEvents =  kCFStreamEventEndEncountered | kCFStreamEventErrorOccurred;

// Notification of URL load
NSString *URLLoadNotification = @"url_load";
NSString *URLLoadSuccessKey = @"url_bool";

@implementation Bookmark

-(id) init
{
  if ((self = [super init])) {
    mUpdateTimer = NULL;
    mURL = [[NSString alloc] init];
    NSNumber *tempNumber = [[NSNumber alloc] initWithUnsignedInt:0];
    mStatus = [tempNumber retain];
    mNumberOfVisits = [tempNumber retain];
    mUpdateTimeIncrement = [tempNumber retain];
    mUpdateTimeUnits = [tempNumber retain];
    mUpdateNotification = [tempNumber retain];
    [tempNumber release];
    mLastVisit = [[NSDate date] retain];
    mIcon = [[NSImage imageNamed:@"smallbookmark"] retain];
    // register for notifications
    NSNotificationCenter *nc = [NSNotificationCenter defaultCenter];
    [nc addObserver:self selector:@selector(siteIconCheck:) name:SiteIconLoadNotificationName object:nil];
    [nc addObserver:self selector:@selector(urlLoadCheck:) name:URLLoadNotification object:nil];
  }
  return self;
}

-(void) initializeLock
{
  if (!mLock)
    mLock = [[NSLock alloc] init];
}

-(id) copyWithZone:(NSZone *)zone
{
  id doppleganger = [super copyWithZone:zone];
  [doppleganger setUrl:[self url]];
  [doppleganger setStatus:[self status]];
  [doppleganger setLastVisit:[self lastVisit]];
  [doppleganger setNumberOfVisits:[self numberOfVisits]];
  [doppleganger setUpdateTimeIncrement:[self updateTimeIncrement]];
  [doppleganger setUpdateTimeUnits:[self updateTimeUnits]];
  [doppleganger setUpdateNotification:[self updateNotification]];
  return doppleganger;
}

- (void) dealloc
{
  [[NSNotificationCenter defaultCenter] removeObserver:self];
  mParent = NULL;	// not retained, so just set to null
  [self setUpdateTimer:NULL];
  if (mLock) {
    [mLock release];
    mLock = nil;
  }
  [mURL release];
  [mLastVisit release];
  [mStatus release];
  [mNumberOfVisits release];
  [mUpdateTimeIncrement release];
  [mUpdateTimeUnits release];
  [mUpdateNotification release];
  [super dealloc];
}

// set/get properties

// url & last visit may be used in other threads.
// just to be safe, we'll return copies of their values
// rather than the actual value
-(NSString *) url
{
  NSString *aString;
  [mLock lock];
  aString = [mURL retain];
  [mLock unlock];
  return [aString autorelease];
}

-(NSDate *) lastVisit
{
  NSDate *aDate;
  [mLock lock];
  aDate = [mLastVisit retain];
  [mLock unlock];
  return [aDate autorelease];
}

-(unsigned) status
{
  return [mStatus unsignedIntValue];
}

-(unsigned) numberOfVisits
{
  return [mNumberOfVisits unsignedIntValue];
}

-(unsigned) updateTimeIncrement
{
  return [mUpdateTimeIncrement unsignedIntValue];
}

-(unsigned) updateTimeUnits
{
  return [mUpdateTimeUnits unsignedIntValue];
}

-(unsigned) updateNotification
{
  return [mUpdateNotification unsignedIntValue];
}

-(NSTimer *)updateTimer
{
  return mUpdateTimer;
}

-(BOOL)	isSeparator
{
  if ([self status]==kBookmarkSpacerStatus)
    return YES;
  return NO;
}

-(BOOL) isUpdated
{
  if ([self status]==kBookmarkUpdatedStatus)
    return YES;
  return NO;
}

-(BOOL) isSick
{
  if (([self status]==kBookmarkBrokenLinkStatus) || ([self status]==kBookmarkServerErrorStatus) || ([self status]==kBookmarkMovedLinkStatus))
    return YES;
  return NO;
}

-(BOOL) iconOnUpdate
{
  if (([self updateNotification] & kNotificationIconOnUpdate) != 0)
    return YES;
  return NO;
}

-(BOOL) beepOnUpdate
{
  if (([self updateNotification] & kNotificationBeepOnUpdate) != 0)
    return YES;
  return NO;
}


- (void) setParent:(id)aParent
{
  [super setParent:aParent];
  if (!aParent)
    [self setUpdateTimer:nil];
}

-(void) setStatusWithNSNumber:(NSNumber *)aNumber
{
  if (aNumber)
    [self setStatus:[aNumber unsignedIntValue]];
}

//
// this gets called on background thread only once - when
// bookmarks first loaded.  so don't need to worry about making
// sure things get called on main thread
//

-(void) setStatus:(unsigned)aStatus
{
  if ((aStatus != [mStatus unsignedIntValue]) &&
      ((aStatus == kBookmarkUpdatedStatus) ||
       (aStatus == kBookmarkBrokenLinkStatus) ||
       (aStatus == kBookmarkMovedLinkStatus) ||
       (aStatus == kBookmarkOKStatus) ||
       (aStatus == kBookmarkSpacerStatus) ||
       (aStatus == kBookmarkServerErrorStatus)))
  {
    [mStatus release];
    mStatus = [[NSNumber alloc] initWithUnsignedInt:aStatus];
    [self itemUpdatedNote];
    switch (aStatus) {
      case kBookmarkUpdatedStatus:
      case kBookmarkBrokenLinkStatus:
      case kBookmarkMovedLinkStatus:
        [self setUpdateTimer:nil];
        break;
      case kBookmarkOKStatus:
        [self scheduleTimer];
        [self refreshIcon];
        break;
      case kBookmarkServerErrorStatus:
        break;
      case kBookmarkSpacerStatus:
        [self setTitle:NSLocalizedString(@"<Menu Spacer>",@"<Menu Spacer>")];
      default:
        return;
    }
    [self doUpdateNotificationTasks];
  }    
}

- (void) setUrl:(NSString *)aURL
{
  if (!aURL) 
    return;
  [mLock lock];
  BOOL isSame = [mURL isEqualToString:aURL];//avoid refresh icon & schedule timer
  [mLock unlock];
  if (isSame)
    return;
  [aURL retain];
  [mLock lock];
  NSString *curURL = mURL;
  mURL = aURL;
  [mLock unlock];
  [curURL release];
  RunLoopMessenger *messenger = [[BookmarkManager sharedBookmarkManager] mainThreadRunLoopMessenger];
  if (messenger) {
    if (messenger == [RunLoopMessenger runLoopMessengerForCurrentRunLoop]) {
      [self scheduleTimer];
      [self refreshIcon];
      [self setStatus:kBookmarkOKStatus];
    } else {
      [messenger target:self performSelector:@selector(scheduleTimer)];
      [messenger target:self performSelector:@selector(refreshIcon)];
      [messenger target:self performSelector:@selector(setStatusWithNSNumber:) withObjects:[NSArray arrayWithObject:[NSNumber numberWithUnsignedInt:kBookmarkOKStatus]]];
    }
  }
}

- (void) setLastVisit:(NSDate *)aDate
{
  if (aDate) {
    NSDate *curDate;
    [aDate retain];
    [mLock lock];
    curDate = mLastVisit;
    mLastVisit = aDate;
    [mLock unlock];
    [curDate release];
  }
}

-(void) setNumberOfVisits:(unsigned)aNumber
{
  [mNumberOfVisits release];
  mNumberOfVisits = [[NSNumber alloc] initWithUnsignedInt:aNumber];
  [self itemUpdatedNote];
}

- (void) setUpdateTimeUnits:(unsigned) aUnits
{
  if (aUnits == kBookmarkNeverUpdate ||
      aUnits == kBookmarkMinutelyUpdate ||
      aUnits == kBookmarkHourlyUpdate ||
      aUnits == kBookmarkDailyUpdate ||
      aUnits == kBookmarkWeeklyUpdate) {
    [mUpdateTimeUnits release];
    mUpdateTimeUnits = [[NSNumber alloc] initWithUnsignedInt:aUnits];
    RunLoopMessenger *messenger = [[BookmarkManager sharedBookmarkManager] mainThreadRunLoopMessenger];
    [messenger target:self performSelector:@selector(scheduleTimer)];
  }
}

- (void) setUpdateTimeIncrement:(unsigned) aNumber
{
    [mUpdateTimeIncrement release];
    mUpdateTimeIncrement = [[NSNumber alloc] initWithUnsignedInt:aNumber];
    RunLoopMessenger *messenger = [[BookmarkManager sharedBookmarkManager] mainThreadRunLoopMessenger];
    [messenger target:self performSelector:@selector(scheduleTimer)];
}

-(void) setUpdateNotification:(unsigned)aNumber
{
  [mUpdateNotification release];
  mUpdateNotification = [[NSNumber alloc] initWithUnsignedInt:aNumber];
}

- (void) setUpdateTimer:(NSTimer *)aTimer
{
  if (mUpdateTimer != aTimer) {
    if (mUpdateTimer)
      [mUpdateTimer invalidate];
    [aTimer retain];
    [mUpdateTimer release];
    mUpdateTimer = aTimer;
  }
}

-(void) setIsSeparator:(BOOL)aBool
{
  if (aBool)
    [self setStatus:kBookmarkSpacerStatus];
  else
    [self setStatus:kBookmarkOKStatus];
}

-(void) setBeepOnUpdate:(BOOL)aBool
{
  unsigned newVal = [mUpdateNotification unsignedIntValue];
  if (aBool) 
    newVal |= kNotificationBeepOnUpdate;
  else 
    newVal &= ~kNotificationBeepOnUpdate;
  [self setUpdateNotification:newVal];
}

-(void) setIconOnUpdate:(BOOL)aBool
{
  unsigned newVal = [mUpdateNotification unsignedIntValue];
  if (aBool)
    newVal |= kNotificationIconOnUpdate;
  else 
    newVal &= ~kNotificationIconOnUpdate;
  [self setUpdateNotification:newVal];
}

//
// see if our icon should be updated
//
-(void) doUpdateNotificationTasks
{
  if ([self beepOnUpdate])
    NSBeep();
  if ([self iconOnUpdate]) {
    if ([self isUpdated])
      [self setIcon:[NSImage imageNamed:@"updatedbookmark_icon"]];
    else if ([self isSick])
      [self setIcon:[NSImage imageNamed:@"brokenbookmark_icon"]];
  }
}

-(void) refreshIcon
{
  [[SiteIconProvider sharedFavoriteIconProvider] loadFavoriteIcon:self forURI:[self url] allowNetwork:NO];
}

-(void) siteIconCheck:(NSNotification *)note
{
  NSDictionary *userInfo = [note userInfo];
  if (userInfo) {
    id loadTarget = [note object];
    NSString *iconString = [userInfo objectForKey:SiteIconLoadUserDataKey];
    if (iconString) {
      NSURL *iconURL= [NSURL URLWithString:[NSString escapedURLString:iconString]];
      NSURL *myURL = [NSURL URLWithString:[NSString escapedURLString:[self url]]];
      if ((loadTarget == self) || [iconURL isEqual:myURL] || [[[iconURL host] stripWWW] isEqualToString:[[myURL host] stripWWW]]) {
        NSImage *iconImage = [userInfo objectForKey:SiteIconLoadImageKey];
        if (iconImage)
          [self setIcon:iconImage];
      }
    }
  }
}
//
// from loads, we only get success or failure - not why things failed.
// so we'll call everything "server error"
//
-(void) urlLoadCheck:(NSNotification *)note
{
  NSString *loadedURL = [note object];
  if ([loadedURL hasSuffix:@"/"])
    loadedURL = [loadedURL substringToIndex:([loadedURL length]-1)];
  NSString *myURL = [self url];
  if ([myURL hasSuffix:@"/"])
    myURL = [myURL substringToIndex:([myURL length]-1)];
  if ([loadedURL isEqualToString:myURL]) {
    unsigned loadStatus = [[[note userInfo] objectForKey:URLLoadSuccessKey] unsignedIntValue];
    [self setLastVisit:[NSDate date]];
    [self setStatus:loadStatus];
    if (loadStatus == kBookmarkOKStatus) 
      [self setNumberOfVisits:([self numberOfVisits]+1)];
  }
}

#pragma mark -
//
// handles checking for updates & whatnot of itself
//

-(NSURL *)urlAsURL
{
  // We'll just assume if there's a # it's a fragment marker.  Yes,
  // this is almost certainly a bug, but just for checking status so who cares.
  NSString *escapedURL = (NSString *)CFURLCreateStringByAddingPercentEscapes
  (NULL,(CFStringRef)[self url],CFSTR("#"),NULL,kCFStringEncodingUTF8);
  NSURL *myURL = [NSURL URLWithString:escapedURL];
  [escapedURL release];
  if (!myURL) {
    NSLog(@"%@ can't form NSURL, turning off update checking",[self url]);
    [self setUpdateTimeUnits:kBookmarkNeverUpdate];
  }
  return myURL;
}

-(void) scheduleTimer
{
  if ([self updateTimeIncrement] == 0 ||
      ([self status] == kBookmarkUpdatedStatus) ||
      ([self status] == kBookmarkBrokenLinkStatus) ||
      ([self status] == kBookmarkMovedLinkStatus))
    return;
  RunLoopMessenger *messenger = [[BookmarkManager sharedBookmarkManager] backgroundRunLoopMessenger];
  if (messenger) {
    NSTimeInterval updateTime;
    switch ([self updateTimeUnits]) {
      case kBookmarkNeverUpdate:
        [self setUpdateTimer:NULL];
        return;
      case kBookmarkMinutelyUpdate:
        updateTime = [self updateTimeIncrement] * 60.0;
        break;
      case kBookmarkHourlyUpdate:
        updateTime = [self updateTimeIncrement] * 3600.0;
        break;
      case kBookmarkDailyUpdate:
        updateTime = [self updateTimeIncrement] * 86400.0;
        break;
      case kBookmarkWeeklyUpdate:
        updateTime = [self updateTimeIncrement] * 604800.0;
        break;
    }
    [self initializeLock];
    NSTimer *aTimer = [[NSTimer alloc] initWithFireDate:[[self lastVisit] addTimeInterval:updateTime] interval:updateTime target:self selector:@selector(checkForUpdate:) userInfo:nil repeats:YES];
    [messenger target:self performSelector:@selector(startTimerInBackground:) withObjects:[NSArray arrayWithObject:aTimer] withResult:NO];
    [self setUpdateTimer:aTimer];
    [aTimer release];
  }
}

-(void) startTimerInBackground:(NSTimer *)aTimer
{
  [[NSRunLoop currentRunLoop] addTimer:aTimer forMode:NSDefaultRunLoopMode];
}

// for now, we'll only check updates on file or http scheme.
// don't know if https will work.
- (void) checkForUpdate:(NSTimer *)aTimer
{
  NSURL* myURL = [self urlAsURL];
  if (!myURL)
    return;
  RunLoopMessenger *messenger = [[BookmarkManager sharedBookmarkManager] mainThreadRunLoopMessenger];
  if ([[myURL scheme] isEqualToString:@"http"])
    [self doHTTPUpdateRequest:myURL];
  else if ([[myURL scheme] isEqualToString:@"file"]) 
    //NSFileManager not thread safe, so do this on main thread
    [messenger target:self performSelector:@selector(doFileUpdateRequest) withObjects:[NSArray arrayWithObject:myURL]];
  else {
    NSLog(@"%@ not http or file, turning off update checking",myURL);
    [messenger target:self performSelector:@selector(invalidSchemeForUpdating)];
  }
}

-(void) invalidSchemeForUpdating
{
  [self setUpdateTimeUnits:kBookmarkNeverUpdate];
  [self setUpdateTimer:nil];
}

//
// CF Callback functions for handling bookmark updating
//

static void doHTTPUpdateCallBack(CFReadStreamRef stream, CFStreamEventType type, void *bookmark)
{
  CFHTTPMessageRef aResponse = NULL;
  NSString* newURL = NULL;
  NSURL* aURL = NULL;
  CFStreamError anError;
  UInt32 errCode;
  switch (type){

    case kCFStreamEventEndEncountered:
      aResponse = (CFHTTPMessageRef) CFReadStreamCopyProperty(stream,kCFStreamPropertyHTTPResponseHeader);
      if (aResponse) {
        errCode = CFHTTPMessageGetResponseStatusCode(aResponse);
        newURL = (NSString*)CFHTTPMessageCopyHeaderFieldValue(aResponse,CFSTR("Location"));
//               NSLog(@"bookmark = %@",[(Bookmark *)bookmark title]);
//                NSLog(@"error code = %u",errCode);
//                NSLog(@"location = %@",newURL);
//                NSLog(@"entire header = %@",(NSString *)CFHTTPMessageCopyAllHeaderFields(aResponse));
        switch (errCode) {
          case 301: //permanent move - update URL and try again
            [(Bookmark *)bookmark setUrl:newURL];
            [(Bookmark *)bookmark doHTTPUpdateRequest:[NSURL URLWithString:newURL]];
            break;
          case 302: //Found - but don't update URL. re-request with suggested URL
          case 303: //See Other - don't update URL. re-request with suggested URL
          case 307: //Temporary Redirect - don't update URL. re-request with suggested URL
                    // I was gonna use Apple's automatic redirect flag on CFHTTPStreams, but
                    // it seems to bite it at sites which don't provide a fully qualified path
                    // (I'm talking to YOU, MSNBC).  So we'll use this hacky version instead.
            aURL = [[NSURL alloc] initWithString:newURL];
            if (![aURL scheme]) {
              [aURL release];
              aURL = [[NSURL alloc] initWithString:[NSString stringWithFormat:@"http://%@%@", [[(Bookmark *)bookmark urlAsURL] host], newURL]];
            }
            [(Bookmark *)bookmark doHTTPUpdateRequest:aURL];
            [aURL release];
            break;
          default:
            [(Bookmark *)bookmark interpretHTTPUpdateCode:errCode];
            break;
        }
        [newURL release];
      } else //beats me.  blame the server.
        [(Bookmark *)bookmark interpretHTTPUpdateCode:500];
      break;

    case kCFStreamEventErrorOccurred:
      anError = CFReadStreamGetError(stream);
      if (anError.domain == kCFStreamErrorDomainHTTP) {
        errCode = anError.error; //signed being assigned to unsinged.  oh well
        [(Bookmark *)bookmark interpretHTTPUpdateCode:errCode];
      } else // call it server error
        [(Bookmark *)bookmark interpretHTTPUpdateCode:500];
      break;

    default:
      NSLog(@"If you can read this you're too close to the screen.");
      break;
  }
  //
  // update our last visit date & cleanup
  //
  [(Bookmark *)bookmark setLastVisit:[NSDate date]];
  if (aResponse)
    CFRelease(aResponse);
}

// borrowed heavily from Apple's CFNetworkHTTPDownload example
-(void) doHTTPUpdateRequest:(NSURL *)aURL
{
  CFHTTPMessageRef messageRef = NULL;
  CFReadStreamRef readStreamRef	= NULL;
  CFStreamClientContext	ctxt= { 0, (void*)self, NULL, NULL, NULL }; //pointer to self lets us update
  // get message
  messageRef = CFHTTPMessageCreateRequest(kCFAllocatorDefault, CFSTR("HEAD"), (CFURLRef)aURL, kCFHTTPVersion1_1);
  if (!messageRef) {
    NSLog(@"CheckForUpdate: Can't create CFHTTPMessage for %@",aURL);
    return;
  }
  // set if-modified-since header field, and maybe others if we're bored.
  NSString *httpDate = [[self lastVisit] descriptionWithCalendarFormat:@"%a, %d %b %Y %H:%M:%S GMT" timeZone:[NSTimeZone timeZoneWithAbbreviation:@"GMT"] locale:nil];
  CFHTTPMessageSetHeaderFieldValue(messageRef,CFSTR("If-Modified-Since"),(CFStringRef)httpDate);

  //setup read stream
  readStreamRef	= CFReadStreamCreateForHTTPRequest(kCFAllocatorDefault, messageRef);
  if (!readStreamRef) {
    NSLog(@"CheckForUpdate: Can't create CFReadStream for %@",aURL);
    CFRelease(messageRef);
    return;
  }
  CFRelease(messageRef);
 //setup callback function
  if (CFReadStreamSetClient(readStreamRef, kNetworkEvents, doHTTPUpdateCallBack, &ctxt ) == false ) {
    NSLog(@"CheckForUpdate: Can't set CFReadStream callback for %@",aURL);
    CFRelease(readStreamRef);
    return;
  }
  //schedule the stream & open the connection
  CFReadStreamScheduleWithRunLoop(readStreamRef, CFRunLoopGetCurrent(), kCFRunLoopCommonModes);
  if (CFReadStreamOpen(readStreamRef) == false ){
    NSLog(@"CheckForUpdate: Can't open CFReadStream for %@",aURL);
    CFReadStreamSetClient(readStreamRef, NULL, NULL, NULL);
    CFRelease(readStreamRef);
    return;
  }
  //schedule a timeout.  we'll give it, oh, 30 seconds before killing the check
  //this timer is responsible for cleaning up the read stream memory!!!!!!!!!
  [NSTimer scheduledTimerWithTimeInterval:30 target:self selector:@selector(cleanupHTTPCheck:) userInfo:(id)readStreamRef repeats:NO];
}

// My interpretation of what should & shouldn't happen is
// quite possibly incorrect.  Feel free to adjust.
-(void) interpretHTTPUpdateCode:(UInt32) errCode
{
  RunLoopMessenger *messenger = [[BookmarkManager sharedBookmarkManager] mainThreadRunLoopMessenger];
  switch (errCode){
    case 304: //Not modified - do nothing
      [messenger target:self performSelector:@selector(setStatusWithNSNumber:) withObjects:[NSArray arrayWithObject:[NSNumber numberWithUnsignedInt:kBookmarkOKStatus]]];
      break;

    case 200: //OK - bookmark updated
    case 203: //Non-authoritative info - call it same as OK
      [messenger target:self performSelector:@selector(setStatusWithNSNumber:) withObjects:[NSArray arrayWithObject:[NSNumber numberWithUnsignedInt:kBookmarkUpdatedStatus]]];
      break;

    case 300: //multiple choices - not sure what to do so we'll bail here
    case 305: //Use proxy (specified) - should retry request
      [messenger target:self performSelector:@selector(setStatusWithNSNumber:) withObjects:[NSArray arrayWithObject:[NSNumber numberWithUnsignedInt:kBookmarkMovedLinkStatus]]];
      break;

    case 400: //Bad request - we f'd up
    case 403: //Forbidden - not good
    case 404: //Not found - clearly not good
    case 410: //Gone - nah nah nahnah, nah nah nahnah, hey hey hey, etc.
      [messenger target:self performSelector:@selector(setStatusWithNSNumber:) withObjects:[NSArray arrayWithObject:[NSNumber numberWithUnsignedInt:kBookmarkBrokenLinkStatus]]];
      break;


    case 301: //Moved permananently - should be handled in callback 
    case 302: //Found - - should be handled in callback
    case 303: //See other - - should be handled in callback
    case 307: //Temporary redirect - - should be handled in callback
    case 401: //Unauthorized - need to be clever about checking this
    case 402: //Payment required - funk that.
    case 405: //Method not allowed - funk that, too.
    case 406: //Not Acceptable - shouldn't happen, but oh well.
    case 407: //Proxy Authentication Required - need to be cleverer here
    case 411: //Length required - need to be cleverer
    case 413: //Request entity too large - shouldn't happen
    case 414: //Request URI too large - shouldn't happen
    case 415: //Request URI too large - shouldn't happen
      [messenger target:self performSelector:@selector(setStatusWithNSNumber:) withObjects:[NSArray arrayWithObject:[NSNumber numberWithUnsignedInt:kBookmarkNeverUpdate]]];
      break;

    case 500: //Internal server error
    case 501: //Not Implemented
    case 502: //Bad Gateway
    case 503: //Service Unavailable
    case 504: //Gateway Timeout
    case 505: //HTTP Version not supported
      [messenger target:self performSelector:@selector(setStatusWithNSNumber:) withObjects:[NSArray arrayWithObject:[NSNumber numberWithUnsignedInt:kBookmarkServerErrorStatus]]];
      break;

    case 100: //Continue - just ignore
    case 101: //Switching protocols - not that smart yet
    case 201: //Created - shouldn't happen
    case 202: //Accepted - shouldn't happen
    case 204: //No content - shouldn't happen
    case 205: //Reset content - shouldn't happen
    case 206: //Partial content - shouldn't happen for HEAD request
    case 408: //Request timeout - shouldn't happen
    case 409: //Conflict - shouldn't happen
    case 412: //Precondintion failed - shouldn't happen
    case 416: //requested range not satisfiable - shouldn't happen
    case 417: //Expectation failed - shouldn't happen
    default:
      break;
  }
}

// always on main thread b/c NSFileManager is not thread-safe
-(void) doFileUpdateRequest:(NSURL *)aURL
{
  // if it's here, it's got a scheme of file, so we can call path directly
  NSFileManager *fM = [NSFileManager defaultManager];
  NSDictionary *attrib = [fM fileAttributesAtPath:[aURL path] traverseLink:YES];
  if ([[self lastVisit] compare:[attrib objectForKey:NSFileModificationDate]] == NSOrderedAscending)
    [self setStatus:kBookmarkUpdatedStatus];
}

// this function cleans up after our stream request.
// if you get rid of it, we leak memory
-(void) cleanupHTTPCheck:(NSTimer *)aTimer;
{
  CFReadStreamRef stream = (CFReadStreamRef)[aTimer userInfo];
  CFReadStreamSetClient(stream,NULL,NULL,NULL);
  CFRelease(stream);
}

#pragma mark -
//
// for reading/writing from/to disk
//

NSString *BMTitleKey = @"Title";
NSString *BMDescKey = @"Description";
NSString *BMStatusKey = @"Status";
NSString *BMURLKey = @"URL";
NSString *BMKeywordKey = @"Keyword";
NSString *BMLastVisitKey = @"LastVisitedDate";
NSString *BMNumberVisitsKey = @"VisitCount";
NSString *BMUpdateNumberKey = @"UpdateNumber";
NSString *BMUpdateUnitsKey = @"UpdateUnits";
NSString *BMUpdateNotificationKey = @"UpdateNotification";

NSString *SafariURIDictKey = @"URIDictionary";
NSString *SafariBookmarkTitleKey = @"title";
NSString *SafariURLStringKey = @"URLString";

-(BOOL) readNativeDictionary:(NSDictionary *)aDict
{
  [self setTitle:[aDict objectForKey:BMTitleKey]];
  [self setDescription:[aDict objectForKey:BMDescKey]];
  [self setStatusWithNSNumber:[aDict objectForKey:BMStatusKey]];
  [self setKeyword:[aDict objectForKey:BMKeywordKey]];
  [self setUrl:[aDict objectForKey:BMURLKey]];
  [self setLastVisit:[aDict objectForKey:BMLastVisitKey]];
  [self setNumberOfVisits:[[aDict objectForKey:BMNumberVisitsKey] unsignedIntValue]];
  [self setUpdateTimeUnits:[[aDict objectForKey:BMUpdateNumberKey] unsignedIntValue]];
  [self setUpdateTimeIncrement:[[aDict objectForKey:BMUpdateUnitsKey] unsignedIntValue]];
  [self setUpdateNotification:[[aDict objectForKey:BMUpdateNotificationKey] unsignedIntValue]];
  return YES;
}
 
-(BOOL) readSafariDictionary:(NSDictionary *)aDict
{
  NSDictionary *uriDict = [aDict objectForKey:SafariURIDictKey];
  [self setTitle:[uriDict objectForKey:SafariBookmarkTitleKey]];
  [self setUrl:[aDict objectForKey:SafariURLStringKey]];
  return YES;
}

-(BOOL) readCaminoXML:(CFXMLTreeRef)aTreeRef
{
  CFXMLNodeRef myNode;
  CFXMLElementInfo* elementInfoPtr;
  myNode = CFXMLTreeGetNode(aTreeRef);
  if (myNode) {
    // Process our info
    if (CFXMLNodeGetTypeCode(myNode)==kCFXMLNodeTypeElement){
      elementInfoPtr = (CFXMLElementInfo *)CFXMLNodeGetInfoPtr(myNode);
      if (elementInfoPtr) {
        NSDictionary* attribDict = (NSDictionary*)elementInfoPtr->attributes;
        [self setTitle:[[attribDict objectForKey:@"name"] stringByRemovingAmpEscapes]];
        [self setKeyword:[[attribDict objectForKey:@"id"] stringByRemovingAmpEscapes]];
        [self setDescription:[[attribDict objectForKey:@"description"] stringByRemovingAmpEscapes]];
        [self setUrl:[[attribDict objectForKey:@"href"] stringByRemovingAmpEscapes]];
      } else {
        NSLog(@"Bookmark:readCaminoXML - elementInfoPtr null, load failed");
        return NO;
      }
    } else {
      NSLog(@"Bookmark:readCaminoXML - node not kCFXMLNodeTypeElement, load failed");
      return NO;
    }
  } else {
    NSLog(@"Bookmark:readCaminoXML - urk! CFXMLTreeGetNode null, load failed");
    return NO;
  }
  return YES;
}

// for plist in native format
-(NSDictionary *)writeNativeDictionary
{
  NSDictionary* dict;
  if (![self isSeparator]) {
    dict = [NSDictionary dictionaryWithObjectsAndKeys:
    mTitle,BMTitleKey,
    mKeyword,BMKeywordKey,
    [self url],BMURLKey,
    mDescription,BMDescKey ,
    [self lastVisit],BMLastVisitKey,
    mNumberOfVisits,BMNumberVisitsKey,
    mUpdateTimeIncrement,BMUpdateNumberKey,
    mUpdateTimeUnits,BMUpdateUnitsKey,
    mUpdateNotification,BMUpdateNotificationKey,
    mStatus,BMStatusKey,
    nil];
  } else {
    dict = [NSDictionary dictionaryWithObjectsAndKeys:
      mStatus,BMStatusKey,
      nil];
  }
  return dict;
}


-(NSString *)writeHTML:(unsigned int)aPad
{
  NSString*formatString;
  NSMutableString *padString = [NSMutableString string];
  for (unsigned i = 0;i<aPad;i++)
    [padString insertString:@"    " atIndex:0];  
  padString = [@" " stringByPaddingToLength:aPad withString:@" " startingAtIndex:0];
  if ([mDescription length] > 0)
    formatString = [NSString stringWithString:@"%@<DT><A HREF=\"%s\">%s</A>\n%@<DD>%s\n"];
  else
    formatString = [NSString stringWithString:@"%@<DT><A HREF=\"%s\">%s</A>\n%@\n"];
  return [NSString stringWithFormat:formatString,
    padString,
    [[self url] UTF8String],
    [[mTitle stringByAddingAmpEscapes] UTF8String],
    padString,
    [[mDescription stringByAddingAmpEscapes] UTF8String]];
}

//
// Scripting methods
//

- (NSScriptObjectSpecifier *)objectSpecifier
{
  id parent = [self parent];
  NSScriptObjectSpecifier *containerRef = [parent objectSpecifier];
  unsigned index = [[parent childArray] indexOfObjectIdenticalTo:self];
  if (index != NSNotFound) {
    // Man, this took forever to figure out.
    // DHBookmarks always contained in BookmarkFolder - so make
    // sure we set that as the container class description.
    NSScriptClassDescription *aRef = (NSScriptClassDescription*)[NSClassDescription classDescriptionForClass:[BookmarkFolder class]];
    return [[[NSIndexSpecifier allocWithZone:[self zone]] initWithContainerClassDescription:aRef containerSpecifier:containerRef key:@"childArray" index:index] autorelease];
  } else
    return nil;
}

@end 

